abstract class Compte_bancaire {
    private _cumul_interets: number = 0;

    constructor(private readonly _id: string, protected _solde: number) {
    }

    public appliquer_taux_interet() {
        this._cumul_interets = this._solde * (1. + (this.taux_interet() / 100.));
    }

    public cumul_interets() {
        return this._cumul_interets;
    }

    public id() {
        return this._id;
    }

    public mise_a_jour(montant: number): number {
        this._solde += montant;
        return this._solde;
    }

    public solde() {
        return this._solde;
    }

    public abstract taux_interet(): number;


    public compareTo(cb: Compte_bancaire): number {
        return this._solde > cb._solde ? 1 : this._solde < cb._solde ? -1 : 0;
    }
}

class Compte_cheque extends Compte_bancaire {
    constructor(_id: string, _solde: number, private readonly _taux_interet: number, private readonly _seuil: number) {
        super(_id, _solde);
    }

    public static Information() {
        return "Classe des comptes bancaires courants ou \"comptes chèque\" - rémunération à la tête du client !";
    }

    public appliquer_taux_interet() {
        if (this._solde > this._seuil) {
            super.appliquer_taux_interet();
        }
    }

    public taux_interet() {
        return this._taux_interet;
    }

}

abstract class Compte_epargne extends Compte_bancaire {
    protected static _Taux_interet: number;

    constructor(_id: string, _solde: number) {
        super(_id, _solde);
    }

    public taux_interet(): number {
        return Compte_epargne._Taux_interet;
    }
}

/** Simulation of Java static initializer */
function Initialize(target: any) { // Decorator...
    target.Initialize();
}

@Initialize // '"experimentalDecorators": true,'
class Livret_A extends Compte_epargne {
    public static Information() {
        return "Classe des livrets A - rémunération fixée par l'Etat !";
    }

    public static Initialize() {
        Compte_epargne._Taux_interet = 0.5; // 13/02/2020
    }
}

@Initialize // '"experimentalDecorators": true,'
class Compte_epargne_logement extends Compte_epargne {
    public static Information() {
        return "Classe des comptes épargne logement - rémunération ratio du Livret A !";
    }

    public static Initialize() { // Launched by decorator...
        Compte_epargne._Taux_interet = Livret_A._Taux_interet * 2. / 3.; // 2/3 du taux du Livret A...
        // Compte_epargne._Taux_interet = ...; // arrondi au 1/4 point le plus proche (0.5 * 2. / 3. ==== 0.33) -> (0.25)...
    }
}

const main_: Main = () => {
    let cb: Compte_bancaire;
    try {
        cb = new Compte_cheque("cc", 200., 0.02, 100.);
        cb.appliquer_taux_interet();
        console.log("cb.taux_interet(): " + cb.taux_interet() + " cb.cumul_interets(): " + cb.cumul_interets());
    } catch (e) {
        console.log(e);
    }
    try {
        cb = new Compte_epargne_logement("cel", 300.);
        cb.appliquer_taux_interet();
        console.log("cb.taux_interet(): " + cb.taux_interet() + " cb.cumul_interets(): " + cb.cumul_interets());
    } catch (e) {
        console.log(e);
    }
};
main_();
